import numpy as np
import glm
import pygame as pg

class BaseModel1:
    def __init__(self, app, vao_name, tex_id, pos, rot, scale, hitbox):
        self.hitbox = hitbox
        self.app = app
        self.pos = pos
        self.rot = glm.vec3([glm.radians(a) for a in rot])
        self.scale = scale
        self.m_model = self.gmm()
        self.tex_id = tex_id
        self.vao_name = vao_name
        self.vao = app.mesh.vao.vaos[vao_name]
        self.program = self.vao.program
        self.camera = self.app.camera

    def update(self): ...

    def gmm(self):
        m_model = glm.mat4()
        m_model = glm.translate(m_model, self.pos)
        m_model = glm.rotate(m_model, self.rot.x, glm.vec3(1,0,0))
        m_model = glm.rotate(m_model, self.rot.y, glm.vec3(0,1,0))
        m_model = glm.rotate(m_model, self.rot.z, glm.vec3(0,0,1))
        m_model = glm.scale(m_model, self.scale)
        return m_model

    def r(self, rot):
        self.rot = glm.vec3([glm.radians(a) for a in rot])

    def render(self):
        self.update()
        self.vao.render()


class BaseModel:
    def __init__(self, app, vao_name, tex_id, pos, rot, scale, hitbox):
        self.hitbox = hitbox
        self.app = app
        self.pos = pos
        self.rot = glm.vec3([glm.radians(a) for a in rot])
        self.scale = scale
        self.m_model = self.gmm()
        self.tex_id = tex_id
        self.vao = app.mesh.vao.vaos[vao_name]
        self.camera = self.app.camera

    def update(self): ...

    def gmm(self):
        m_model = glm.mat4()
        m_model = glm.translate(m_model, self.pos)
        m_model = glm.rotate(m_model, self.rot.x, glm.vec3(1,0,0))
        m_model = glm.rotate(m_model, self.rot.y, glm.vec3(0,1,0))
        m_model = glm.rotate(m_model, self.rot.z, glm.vec3(0,0,1))
        m_model = glm.scale(m_model, self.scale)
        return m_model

    def r(self, rot):
        self.rot = glm.vec3([glm.radians(a) for a in rot])

    def render(self):
        self.update()
        self.vao.render()

class ExBaseModel(BaseModel):
    def __init__(self, app, vao_name, tex_id, pos, rot, scale, hitbox):
        super().__init__(app, vao_name, tex_id, pos, rot, scale, hitbox)
        self.on_init()

    def update(self):
        self.texture.use(location=0)
        [vao[0].program['camPos'].write(self.app.camera.position) for vao in self.vao.vaos]
        [vao[0].program['view'].write(glm.int8(self.app.camera.view)) for vao in self.vao.vaos]
        [vao[0].program['m_view'].write(self.app.camera.m_view) for vao in self.vao.vaos]
        [vao[0].program['m_model'].write(self.m_model) for vao in self.vao.vaos]

        [vao[1].program['m_proj'].write(self.camera.m_proj) for vao in self.vao.vaos]
        [vao[1].program['m_view_light'].write(self.app.light.m_view_light) for vao in self.vao.vaos]

        [vao[0].program['m_proj'].write(self.app.camera.m_proj) for vao in self.vao.vaos]
        #fog = max(min((glm.distance(self.app.camera.position, self.pos) - 100)/20, 1.0), 0);
        #print(fog)
        #[vao[0].program['fog'].write(glm.float64(fog)) for vao in self.vao.vaos]

    def update_shadow(self):
        [vao[1].program['m_model'].write(self.m_model) for vao in self.vao.vaos]

    def render_shadow(self):
        self.update_shadow()
        [vao[1].render() for vao in self.vao.vaos]

    def on_init(self):
        [vao[0].program['m_view_light'].write(self.app.light.m_view_light) for vao in self.vao.vaos]
        [vao[0].program['u_resolution'].write(glm.vec2(self.app.size)) for vao in self.vao.vaos]

        self.depth_texture = self.app.mesh.texture.textures['dt']
        for vao in self.vao.vaos: vao[0].program['shadowMap'] = 1
        self.depth_texture.use(location=1)

        [vao[1].program['m_proj'].write(self.camera.m_proj) for vao in self.vao.vaos]
        [vao[1].program['m_view_light'].write(self.app.light.m_view_light) for vao in self.vao.vaos]
        [vao[1].program['m_model'].write(self.m_model) for vao in self.vao.vaos]

        self.texture = self.app.mesh.texture.gt(self.tex_id)
        for vao in self.vao.vaos: vao[0].program['u_texture_0'] = 0
        self.texture.use(location=0)

        [vao[0].program['m_proj'].write(self.app.camera.m_proj) for vao in self.vao.vaos]
        [vao[0].program['m_view'].write(self.app.camera.m_view) for vao in self.vao.vaos]
        [vao[0].program['m_model'].write(self.m_model) for vao in self.vao.vaos]
        [vao[0].program['light.position'].write(self.app.light.position) for vao in self.vao.vaos]
        [vao[0].program['light.Ia'].write(self.app.light.Ia) for vao in self.vao.vaos]
        [vao[0].program['light.Id'].write(self.app.light.Id) for vao in self.vao.vaos]
        [vao[0].program['light.Is'].write(self.app.light.Is) for vao in self.vao.vaos]

    def tex(self, name=None, tex=None):
        self.tex_id = name
        self.texture = (tex if tex else self.app.mesh.texture.gt(self.tex_id))
        for vao in self.vao.vaos: vao[0].program['u_texture_0'] = 0

class Cube(ExBaseModel):
    def __init__(self, app, tex_id, pos=(0,0,0), rot=(0,0,0), scale=(1,1,1), vao_name='cube', hitbox=None):
        super().__init__(app, vao_name, tex_id, pos, rot, scale, hitbox)

class UniverseModel(ExBaseModel):
    def __init__(self, app, vao_name, tex_id, pos=(0,0,0), rot=(0,0,0), scale=(1,1,1), hitbox=None):
        super().__init__(app, vao_name, tex_id, pos, rot, scale, hitbox)

class MovingCube(Cube):
    def __init__(self, app, tex_id, pos=(0, 0, 0), rot=(0, 0, 0), scale=(1, 1, 1), vao_name='cube', II=None,
                 hitbox=None):
        super().__init__(app, tex_id, pos, rot, scale, vao_name, hitbox)
        if II:
            self.ii = II(self)

    def update(self):
        self.m_model = self.gmm()
        super().update()

class MovingUniverseModel(UniverseModel):
    def __init__(self, app, vao_name, tex_id, pos=(0,0,0), rot=(0,0,0), scale=(1,1,1),II=None, hitbox=None):
        super().__init__(app, vao_name, tex_id, pos, rot, scale, hitbox)
        if II: self.ii = II(self)

    def update(self):
        self.m_model = self.gmm()
        super().update()

class SkyBox(BaseModel1):
    def __init__(self, app, vao_name='skybox', tex_id='1',
                 pos=(0, 0, 0), rot=(0, 0, 0), scale=(1, 1, 1), hitbox=None):
        super().__init__(app, vao_name, tex_id, pos, rot, scale, hitbox)
        self.on_init()

    def update(self):
        self.program['m_view'].write(glm.mat4(glm.mat3(self.camera.m_view)))

    def on_init(self):
        # texture
        self.texture = self.app.mesh.texture.skyBoxes[self.tex_id]
        self.program['u_texture_skybox'] = 0
        self.texture.use()
        # mvp
        self.program['m_proj'].write(self.camera.m_proj)
        self.program['m_view'].write(glm.mat4(glm.mat3(self.camera.m_view)))

class AdvancedSkyBox(BaseModel1):
    def __init__(self, app, tex_id, vao_name='advanced_skybox', pos=(0, 0, 0), rot=(0, 0, 0), scale=(1, 1, 1), hitbox=None):
        super().__init__(app, vao_name, tex_id, pos, rot, scale, hitbox)
        self.on_init()

    def update(self):
        m_view = glm.mat4(glm.mat3(self.camera.m_view))
        self.program['m_invProjView'].write(glm.inverse(self.camera.m_proj * m_view))

    def on_init(self):
        self.texture = self.app.mesh.texture.skyBoxes[self.tex_id]
        self.program['u_texture_skybox'] = 0
        self.texture.use(location=0)

    def tex(self, name):
        self.tex_id = name
        self.texture = self.app.mesh.texture.skyBoxes[self.tex_id]
        self.program['u_texture_skybox'] = 0
        self.texture.use(location=0)



import numpy as np
import moderngl as mgl
from pathlib import Path
import pywavefront
import os

system = os.name

class VBO:
    def __init__(self, ctx):
        print('load models')
        self.folder = Path('accets/models')
        self.vbos = {}
        self.vbos1 = {}
        self.vbos['cube'] = CubeVBO(ctx)
        self.vbos1['skybox'] = SkyBoxVBO(ctx)
        self.vbos1['advanced_skybox'] = AdvancedSkyBoxVBO(ctx)
        for file in self.folder.iterdir():
            file = str(file)
            name = file.split("\\" if system == 'nt' else '/' if system == 'posix' else '')[-1]
            print(name)
            self.vbos[name] = ObjModel(ctx, file)
    def destroy(self):
        [vbo.destroy() for vbo in self.vbos.values()]


class BaseVBO:
    def __init__(self, ctx):
        self.ctx = ctx
        self.vbo = self.get_vbo()
        self.format: str = None
        self.attribs: list = None

    def get_vertex_data(self): ...

    def get_vbo(self):
        vertex_data = self.get_vertex_data()
        vbo = self.ctx.buffer(vertex_data)
        return vbo

    def destroy(self):
        self.vbo.release()


class CubeVBO():
    def __init__(self, ctx):
        self.ctx = ctx
        self.vbos = [self.get_vbo()]
        self.format = '2f 3f 3f'
        self.attribs = ['in_texcoord_0', 'in_normal', 'in_position']

    @staticmethod
    def get_data(vertices, indices):
        data = [vertices[ind] for triangle in indices for ind in triangle]
        return np.array(data, dtype='f4')

    def get_vbo(self):
        vertex_data = self.get_vertex_data()
        vbo = self.ctx.buffer(vertex_data)
        return vbo

    def get_vertex_data(self):
        vertices = [(-1, -1, 1), ( 1, -1,  1), (1,  1,  1), (-1, 1,  1),
                    (-1, 1, -1), (-1, -1, -1), (1, -1, -1), ( 1, 1, -1)]

        indices = [(0, 2, 3), (0, 1, 2),
                   (1, 7, 2), (1, 6, 7),
                   (6, 5, 4), (4, 7, 6),
                   (3, 4, 5), (3, 5, 0),
                   (3, 7, 4), (3, 2, 7),
                   (0, 6, 1), (0, 5, 6)]
        vertex_data = self.get_data(vertices, indices)

        tex_coord_vertices = [(0, 0), (1, 0), (1, 1), (0, 1)]
        tex_coord_indices = [(0, 2, 3), (0, 1, 2),
                             (0, 2, 3), (0, 1, 2),
                             (0, 1, 2), (2, 3, 0),
                             (2, 3, 0), (2, 0, 1),
                             (0, 2, 3), (0, 1, 2),
                             (3, 1, 2), (3, 0, 1),]
        tex_coord_data = self.get_data(tex_coord_vertices, tex_coord_indices)

        normals = [( 0, 0, 1) * 6,
                   ( 1, 0, 0) * 6,
                   ( 0, 0,-1) * 6,
                   (-1, 0, 0) * 6,
                   ( 0, 1, 0) * 6,
                   ( 0,-1, 0) * 6,]
        normals = np.array(normals, dtype='f4').reshape(36, 3)

        vertex_data = np.hstack([normals, vertex_data])
        vertex_data = np.hstack([tex_coord_data, vertex_data])
        return vertex_data

    def destroy(self): self.vbos[0].release()


class ObjModel:
    def __init__(self, app, path):
        self.path = path
        self.ctx = app
        self.vbos = []
        self.vbo = self.get_vbo()
        self.format = '2f 3f 3f'
        self.attribs = ['in_texcoord_0', 'in_normal', 'in_position']
    def get_vbo(self):
        objs = pywavefront.Wavefront(self.path + ('\obj.obj' if system == 'nt' else '/obj.obj' if system == 'posix' else ''), cache=True, parse=True, strict=False)
        for name, obj in objs.materials.items(): self.vbos.append(self.ctx.buffer(np.array(obj.vertices, dtype='f4')))

    def destroy(self):
        [vbo.release() for vbo in self.vbos]


class SkyBoxVBO(BaseVBO):
    def __init__(self, ctx):
        super().__init__(ctx)
        self.format = '3f'
        self.attribs = ['in_position']

    @staticmethod
    def get_data(vertices, indices):
        data = [vertices[ind] for triangle in indices for ind in triangle]
        return np.array(data, dtype='f4')

    def get_vertex_data(self):
        vertices = [(-1, -1, 1), ( 1, -1,  1), (1,  1,  1), (-1, 1,  1),
                    (-1, 1, -1), (-1, -1, -1), (1, -1, -1), ( 1, 1, -1)]

        indices = [(0, 2, 3), (0, 1, 2),
                   (1, 7, 2), (1, 6, 7),
                   (6, 5, 4), (4, 7, 6),
                   (3, 4, 5), (3, 5, 0),
                   (3, 7, 4), (3, 2, 7),
                   (0, 6, 1), (0, 5, 6)]
        vertex_data = self.get_data(vertices, indices)
        vertex_data = np.flip(vertex_data, 1).copy(order='C')
        return vertex_data


class AdvancedSkyBoxVBO(BaseVBO):
    def __init__(self, ctx):
        super().__init__(ctx)
        self.format = '3f'
        self.attribs = ['in_position']

    def get_vertex_data(self):
        # in clip space
        z = 0.9999
        vertices = [(-1, -1, z), (3, -1, z), (-1, 3, z)]
        vertex_data = np.array(vertices, dtype='f4')
        return vertex_data


import glm


class Light:
    def __init__(self, position=(50, 50, -10), color=(0.86, 0.8, 0.9)):
        self.position = glm.vec3(position)#(0.86, 0.8, 0.9)
        self.color = glm.vec3(color)
        self.direction = glm.vec3(0, 0, 0)

        self.Ia = 0.03 * self.color
        self.Id = 0.8 * self.color
        self.Is = 7.0 * self.color

        self.m_view_light = self.get_view_matrix()

    def get_view_matrix(self):
        return glm.lookAt(self.position, self.direction, glm.vec3(0, 1, 0))



import math
import glm


def sphInter(p1, p2, rd, ra):
    ro = p1-p2
    b = glm.dot(ro, rd)
    c = glm.dot(ro, ro) - ra * ra
    h = b * b - c
    it = None
    if h < 0.0: it = glm.vec2(-1.0, -1.0)
    else: h = math.sqrt(h); it = glm.vec2(-b - h, -b + h);
    it = it[0] if it[0] > 0.0 else it[1]
    if it < 0.0: return False
    else: itPos = p1 + rd * it; return glm.distance(p1, itPos)

def sumM(lists):
    list1 = [0 for i in range(len(lists[0]))]
    for _ in lists:
        for i in range(len(_)):
            list1[i] += _[i]
    return list1

def revList(list):
    return [-i for i in list]

def rotPoint(p, p1, alpha, version='2d'):
    if version == '2d':
        X = p1.x + ((p.x-p1.x)*math.cos(math.radians(alpha)) - (p.y-p1.y)*math.sin(math.radians(alpha)))
        Y = p1.y + ((p.x-p1.x)*math.sin(math.radians(alpha)) + (p.y-p1.y)*math.cos(math.radians(alpha)))
        return glm.vec2(X, Y)
    elif version == '3d':
        X = p.x
        Z = p1.z + ((p.z - p1.z) * math.cos(math.radians(alpha[0])) - (p.y - p1.y) * math.sin(math.radians(alpha[0])))
        Y = p1.y + ((p.z - p1.z) * math.sin(math.radians(alpha[0])) + (p.y - p1.y) * math.cos(math.radians(alpha[0])))

        X = p1.x + ((Z - p1.z) * math.sin(math.radians(alpha[1])) + (X - p1.x) * math.cos(math.radians(alpha[1])))
        Z = p1.z + ((Z - p1.z) * math.cos(math.radians(alpha[1])) - (X - p1.x) * math.sin(math.radians(alpha[1])))
        Y = Y

        X = p1.x + ((Y - p1.y) * math.sin(math.radians(alpha[2])) + (X - p1.x) * math.cos(math.radians(alpha[2])))
        Y = p1.y + ((Y - p1.y) * math.cos(math.radians(alpha[2])) - (X - p1.x) * math.sin(math.radians(alpha[2])))
        Z = Z
        return glm.vec3(X, Y, Z)

def rotFigure(poss, pos, alpha, version='2d'):
    poss1 = []
    for p in poss:
        poss1.append(rotPoint(p, pos, alpha, version))
    return poss1

def isIn(a, b, c, version='1d'):
    if version == '1d':
        return (a >= b and a <= c)
    elif version == '2d':
        return (a.x >= min(b.x, c.x) and a.x <= max(b.x, c.x) and a.y >= min(b.y, c.y) and a.y <= max(b.y, c.y))
    elif version == '3d':
        return (a.x >= min(b.x, c.x) and a.x <= max(b.x, c.x) and a.y >= min(b.y, c.y) and a.y <= max(b.y, c.y) and a.z >= min(b.z, c.z) and a.z <= max(b.z, c.z))

def checkCollizion(rect, rect1, version='2d'):
    col = False
    for p in rect1.gposs:
        rp = rotPoint(rotPoint(p, rect1.pos, rect1.angle, version), rect.pos, (-rect.angle if version == '2d' else revList(rect.angle)), version)
        if isIn(rp, rect.grect[0], rect.grect[1], version): col = True
    for p in rect.gposs:
        rp = rotPoint(rotPoint(p, rect.pos, rect.angle, version), rect1.pos, (-rect1.angle if version == '2d' else revList(rect1.angle)), version)
        if isIn(rp, rect1.grect[0], rect1.grect[1], version): col = True
    return col

class Rect:
    def __init__(self, pos, rect):
        self.pos = pos
        self.lrect = rect
        self.grect = list(map(lambda x: x + self.pos, self.lrect))
        self.lposs = [self.lrect[0],
                      glm.vec2(list(self.lrect[0])[0], list(self.lrect[1])[1]),
                      self.lrect[1],
                      glm.vec2(list(self.lrect[1])[0], list(self.lrect[0])[1])]
        self.gposs = list(map(lambda x: x + self.pos, self.lposs))
        self.angle = 0

class Cube:
    def __init__(self, pos, cube):
        self.pos = pos
        self.lrect = cube
        self.grect = list(map(lambda x: x + self.pos, self.lrect))
        self.lposs = [self.lrect[0],
                      glm.vec3(list(self.lrect[0])[0], list(self.lrect[0])[1], list(self.lrect[1])[2]),
                      glm.vec3(list(self.lrect[0])[0], list(self.lrect[0])[2], list(self.lrect[1])[0]),
                      glm.vec3(list(self.lrect[0])[0], list(self.lrect[0])[2], list(self.lrect[1])[1]),
                      glm.vec3(list(self.lrect[1])[0], list(self.lrect[1])[1], list(self.lrect[0])[2]),
                      glm.vec3(list(self.lrect[1])[0], list(self.lrect[1])[2], list(self.lrect[0])[0]),
                      glm.vec3(list(self.lrect[1])[0], list(self.lrect[1])[2], list(self.lrect[0])[1]),
                      self.lrect[1]]
        self.gposs = list(map(lambda x: x + self.pos, self.lposs))
        self.angle = [0, 0, 0]

    def update(self, pos):
        self.pos = pos
        self.grect = list(map(lambda x: x + pos, self.lrect))
        self.gposs = list(map(lambda x: x + pos, self.lposs))



# boxN;
# vec3
# boxPos = vec3(3.0, 1.0, -0.001);
# it = boxIntersection(ro - boxPos, rd, vec3(1.0), boxN);
# if (it.x > 0 & & it.x < it.y) {it.x = it.x;} else {it.x = it.y;};
# if (it.x > 0.0 & & it.x < minIt.x) {
# minIt = it;
# n = boxN;
# col = vec4(0.9, 0.2, 0.2, -0.5);
# }



import moderngl as mgl
import pygame as pg
import numpy as np
from moviepy.editor import VideoFileClip
import cv2, os, time
from threading import Thread as Th
import sounddevice as sd

class Clip:
    def __init__(self, path, end):
        self.pause = True
        self.pTime = time.time()
        self.img = pg.Surface((100, 100))
        self.end = end
        self.col = 0
        if path != 0:
            self.path = path
            self.video = VideoFileClip(path)
            self.music = self.video.audio.to_soundarray()
            self.fps = self.video.fps
            #sd.play(self.music, 48000)
        else:
            self.path = path
            self.fps = 20
        self.cap = cv2.VideoCapture(path)
        if self.cap.isOpened() == False: print(f'Не возможно открыть файл {name}')

    def run(self):
        while True:
            self.update()

    def update(self):
        if self.pause: self.pTime = time.time()
        l = time.time() - self.pTime
        n = round(l / (1/self.fps))
        fl, img = None, None
        for i in range(n):
            fl, img = self.cap.read()
            if img is None and self.path != 0: self.end(self)
        if not fl: return
        self.img = pg.image.frombuffer(img.tobytes(), img.shape[1::-1], "BGR")
        self.pTime = time.time()

    def step(self, path, end=lambda: 0):
        self.release()
        self.__init__(path, end)

    def release(self):
        self.cap.release()

class System:
    def __init__(self, app):
        self.os = os.name
        self.app = app
        self.clips = {}

    def loadClipPack(self, datas):
        for data in datas:
            texPath = ''
            if data[0] != 0:
                if self.os == 'posix': texPath = f'./accets/clips/{data[0]}'
                elif self.os == 'nt': texPath = f'accets\\clips\\{data[0]}'
            else: texPath = 0
            self.clips[data[1]] = Clip(texPath, data[2])

    def release(self):
        [clip.release() for clip in self.clips.values()]


from .vao import VAO
from .texture import Texture


class Mesh:
    def __init__(self, app):
        self.app = app
        self.vao = VAO(app.ctx)
        self.texture = Texture(app)

    def destroy(self):
        self.vao.destroy()
        self.texture.destroy()


from .vbo import VBO
from .shader_program import ShaderProgram

class Vao:
    def __init__(self, progs, vbo, ctx):
        self.ctx = ctx
        self.vbo = vbo
        self.vaos = [[self.get_vao(program=progs['default'], vbo=v, v=vbo), self.get_vao(program=progs['shadow_map'], vbo=v, v=vbo)] for v in vbo.vbos]

    def get_vao(self, program, vbo, v):
        vao = self.ctx.vertex_array(program, [(vbo, v.format, *v.attribs)], skip_errors=True)
        return vao

    def render(self):
        [vao[0].render() for vao in self.vaos]

class VAO:
    def __init__(self, ctx):
        self.ctx = ctx
        self.vbo = VBO(ctx)
        self.program = ShaderProgram(ctx)
        self.vaos = {}
        for name in self.vbo.vbos: self.vaos[name] = Vao(self.program.programs, self.vbo.vbos[name], ctx)
        self.vaos['cube'] = Vao(self.program.programs, self.vbo.vbos['cube'], ctx)
        self.vaos['skybox'] = self.get_vao(program=self.program.programs['skybox'], vbo=self.vbo.vbos1['skybox'])
        self.vaos['advanced_skybox'] = self.get_vao(program=self.program.programs['advanced_skybox'],vbo=self.vbo.vbos1['advanced_skybox'])

    def get_vao(self, program, vbo):
        vao = self.ctx.vertex_array(program, [(vbo.vbo, vbo.format, *vbo.attribs)], skip_errors=True)
        return vao

    def destroy(self):
        self.vbo.destroy()
        self.program.destroy()



from . import render
# from . import phisics

class GameEngine:
    def __init__(self, size, tp):
        self.rend = render.GraphicsEngine(size=size)
        self.tp = tp(self.rend, self)
        self.rend.defDr = self.tp.render

    def run(self):
        self.tp.start()
        while True:
            self.tp.update()
            self.rend.update()


from . import render
from . import model
from . import colliders
from . import phisics
from . import vidSys
from . import createNewGame

class GameEngine:
    def __init__(self, tp, size=None):
        self.rend = render.GraphicsEngine(size)
        self.tp = tp(self.rend, self)
        self.rend.defDr = self.tp.render
        self.rend.defEv = self.tp.ge
        self.on = True

    def run(self):
        self.tp.start()
        while self.on:
            self.rend.update()

#try: except: pass


import glm
import pygame as pg
import moderngl as mgl
from array import array
import time, math
from . import model
from .camera import Camera
from .light import Light
from .mesh import Mesh
from .model import *
import pyautogui
size = pyautogui.size()
w, h = size[0], size[1]


def cos(a):
    return math.cos(math.radians(a))


def sin(a):
    return math.sin(math.radians(a))


def tg(a):
    return math.tan(math.radians(a))


def li(line1, line2):
    xdiff = (line1[0][0] - line1[1][0], line2[0][0] - line2[1][0])
    ydiff = (line1[0][1] - line1[1][1], line2[0][1] - line2[1][1])

    def det(a, b):
        return a[0] * b[1] - a[1] * b[0]

    div = det(xdiff, ydiff)
    if div == 0:
        raise Exception('lines do not intersect')

    d = (det(*line1), det(*line2))
    x = det(d, xdiff) / div
    y = det(d, ydiff) / div
    return x, y

def l(x, y):
    return ((x[0] - y[0]) ** 2 + (x[1] - y[1]) ** 2) ** 0.5


class GraphicsEngine:
    def __init__(self, size=None):
        self.full = bool(size)
        self.size = (size if self.full else (w, h))
        print(self.full, size, self.size)
        self.init()
        self.defDr = lambda surf: 0
        self.defEv = lambda: 0
        self.time = 0
        self.delta_time = 0
        self.fps = 1
        self.mesh = Mesh(self)
        self.light = Light()
        self.scene = self.Scene(self)
        self.validate = False
        self.clock = pg.time.Clock()

    class Scene:
        def __init__(self, app):
            self.scene = {"objects":{}, "render":[], "shadow":[], "II":[]}
            self.scs = {}
            self.app = app
            self.ctx = app.ctx
            self.mesh = app.mesh
            self.sky = model.AdvancedSkyBox(self.app, 'default')
            self.name = None
            self.depth_texture = self.mesh.texture.textures['dt']
            self.depth_fbo = self.ctx.framebuffer(depth_attachment=self.depth_texture)

        def d1(self, _):
            _.render()

        def d2(self, _):
            _.render_shadow()

        def add(self, scene, name, obj, render, shadow, II):
            self.scs[scene]['objects'][name] = obj
            if render: self.scs[scene]['render'].append(name)
            if shadow: self.scs[scene]['shadow'].append(name)
            if II: self.scs[scene]['II'].append(name)

        def render(self):
            if self.name:
                if self.sky: self.sky.render()

                [self.scs[self.name]['objects'][_].ii.update() for _ in self.scs[self.name]['II']]

                self.depth_fbo.clear()
                self.depth_fbo.use()
                [self.scs[self.name]['objects'][_].render_shadow() for _ in self.scs[self.name]['shadow']]

                self.app.ctx.screen.use()
                [self.scs[self.name]['objects'][_].render() for _ in self.scs[self.name]['render']]


        def destroy(self): self.depth_fbo.release()

    class Pointer():
        def __init__(self, cam, pos):
            self.cam = cam
            self.pos = pos
            self.fov = cam.FOV
            self.nfov = self.fov//cam.aspect_ratio
            self.angle = [0, 0]
            self.fa = 0

        def behind(self):
            toOther = self.pos - self.cam.position
            return glm.dot(self.cam.forward, toOther) < 0

        def update(self):
            pos = self.cam.position
            self.fa = math.degrees(math.atan2(self.pos.y - pos.y, self.pos.x - pos.x))
            self.fa += self.cam.yaw
            self.fb = math.degrees(math.atan2(self.pos.y - pos.y, self.pos.z - pos.z))
            self.fb += self.cam.pitch
            posn = [li(((cos(-self.fov), sin(-self.fov)), (cos(0), sin(0))), ((0, 0), (cos(-self.fa), sin(-self.fa)))), li(((cos(-self.nfov), sin(-self.nfov)), (cos(0), sin(0))), ((0, 0), (cos(-self.fb), sin(-self.fb))))]
            m = l((cos(-self.fov), sin(-self.fov)), posn[0]) / l((cos(-self.fov), sin(-self.fov)), (cos(0), sin(0)))
            m1 = l((cos(-self.nfov), sin(-self.nfov)), posn[1]) / l((cos(-self.nfov), sin(-self.nfov)), (cos(0), sin(0)))
            self.angle = [m, m1]

    def quit(self):
        self.mesh.destroy()
        self.scene.destroy()
        pg.quit()
        exit()

    def init(self):
        pg.init()

        pg.display.gl_set_attribute(pg.GL_CONTEXT_MAJOR_VERSION, 3)
        pg.display.gl_set_attribute(pg.GL_CONTEXT_MINOR_VERSION, 3)
        pg.display.gl_set_attribute(pg.GL_CONTEXT_PROFILE_MASK, pg.GL_CONTEXT_PROFILE_CORE)

        pg.display.set_mode(self.size, flags=pg.OPENGL | pg.DOUBLEBUF) if self.full else pg.display.set_mode(self.size, flags=pg.OPENGL | pg.DOUBLEBUF | pg.FULLSCREEN)
        self.ctx = mgl.create_context()
        #self.ctx.front_face = 'cw'
        self.ctx.enable(flags=mgl.DEPTH_TEST | mgl.CULL_FACE)

        self.clock = pg.time.Clock()
        self.camera = Camera(self)

        self.surf = pg.Surface(self.size, flags=pg.SRCALPHA)
        self.pg_texture = self.ctx.texture(self.size, 4)
        self.pg_texture.filter = mgl.NEAREST, mgl.NEAREST
        self.pg_texture.swizzle = 'BGRA'
        self.texture_program = self.ctx.program(
            vertex_shader="""
                #version 330
                in vec2 in_vert;
                in vec2 in_texcoord;
                out vec2 uv;
                void main() {
                    uv = in_texcoord;
                    gl_Position = vec4(in_vert, 0.0, 1.0);
                }
            """,
            fragment_shader="""
                #version 330
                uniform sampler2D surface;
                out vec4 f_color;
                in vec2 uv;
                void main() {
                    f_color = texture(surface, uv);
                }
            """)
        self.texture_program['surface'] = 0
        buffer = self.ctx.buffer(
            data=array('f', [
                -1.0, 1.0, 0.0, 1.0,
                -1.0, -1.0, 0.0, 0.0,
                1.0, 1.0, 1.0, 1.0,
                1.0, -1.0, 1.0, 0.0]))
        self.quad_fs = self.ctx.vertex_array(self.texture_program, [(buffer, "2f 2f", "in_vert", "in_texcoord")])

    def render(self):
        self.ctx.clear(0,255,0)
        self.scene.render()
        self.ctx.disable(mgl.BLEND)
        self.defDr(self.surf)
        self.surf = pg.transform.flip(self.surf, flip_x=False, flip_y=True)
        texture_data = self.surf.get_view('1')
        self.pg_texture.write(texture_data)
        self.ctx.enable(mgl.BLEND)
        self.pg_texture.use(location=0)
        self.quad_fs.render(mode=mgl.TRIANGLE_STRIP)
        pg.display.flip()

    def update(self):
        nTime = time.time()
        self.defEv()
        self.render()
        self.fps = 1 / (time.time()-nTime) if (time.time()-nTime) != 0 else 0.00001




class ShaderProgram:
    def __init__(self, ctx):
        self.ctx = ctx
        self.programs = {}
        self.programs['default'] = self.get_program('default')
        self.programs['skybox'] = self.get_program('skybox')
        self.programs['advanced_skybox'] = self.get_program('advanced_skybox')
        self.programs['shadow_map'] = self.get_program('shadow_map')

    def get_program(self, shader_program_name):
        with open(f'accets/shaders/{shader_program_name}.vert') as file:
            vertex_shader = file.read()

        with open(f'accets/shaders/{shader_program_name}.frag') as file:
            fragment_shader = file.read()

        program = self.ctx.program(vertex_shader=vertex_shader, fragment_shader=fragment_shader)
        return program

    def destroy(self):
        [program.release() for program in self.programs.values()]


from setuptools import setup

setup(name='akira3d',
      version='4.8.1',
      description='Modern 3D Graphics Engine by alexCoder23',
      packages=['akira3d'],
      author_email='alekseybeldem@gmail.com',
      zip_safe=False)



import moderngl as mgl
import pygame as pg
import numpy as np
import os

class Texture:
    def __init__(self, app):
        self.os = os.name
        self.app = app
        self.ctx = app.ctx
        self.textures = {'default':self.get_texture('default.png'), 'dt':self.get_depth_texture()}
        self.skyBoxes = {'default':self.get_texture_cube('default')}

    def loadTexPack(self, names):
        for name in names:
            self.textures[name] = self.get_texture(name)

    def loadSkyPack(self, names):
        for name in names:
            self.skyBoxes[name] = self.get_texture_cube(name)

    def get_depth_texture(self):
        tex = self.ctx.depth_texture(self.app.size)
        tex.repeat_x = False
        tex.repeat_y = False
        return tex

    def get_texture_cube(self, name):
        faces = ['right', 'left', 'top', 'bottom'] + ['front', 'back'][::-1]
        textures = []
        for face in faces:
            texPath = ''
            if self.os == 'posix':
                texPath = f'./accets/skyboxes/{name}/{face}.png'
            elif self.os == 'nt':
                texPath = f'accets\\skyboxes\\{name}\\{face}.png'
            texture = pg.image.load(texPath).convert_alpha()
            if face in ['right', 'left', 'front', 'back']:
                texture = pg.transform.flip(texture, flip_x=True, flip_y=False)
            else:
                texture = pg.transform.flip(texture, flip_x=False, flip_y=True)
            textures.append(texture)
        size = textures[0].get_size()
        texture_cube = self.ctx.texture_cube(size=size, components=3, data=None)
        for i in range(6):
            texture_data = pg.image.tostring(textures[i], 'RGB')
            texture_cube.write(face=i, data=texture_data)
        return texture_cube

    def get_texture(self, name):
        texPath = ''
        if self.os == 'posix':
            texPath = f'./accets/textures/{name}'
        elif self.os == 'nt':
            texPath = f'accets\\textures\\{name}'
        img = pg.image.load(texPath).convert_alpha()
        return self.create_texture(img)

    def create_texture(self, img):
        img = pg.transform.flip(img, flip_x=False, flip_y=True)
        texture = self.ctx.texture(size=img.get_size(), components=4, data=pg.image.tostring(img, 'RGBA'))
        texture.filter = (mgl.LINEAR_MIPMAP_LINEAR, mgl.LINEAR)
        texture.swizzle = 'RGBA'
        texture.build_mipmaps()
        texture.anisotropy = 32.0
        return texture

    def gt(self, name):
        if name in self.textures.keys(): return self.textures[name]
        else: return self.textures['default']

    def destroy(self):
        [tex.release() for tex in self.textures.values()]



from glm import *
import math

class Mob:
    def __init__(self):
        self.bodyParts = {}

    def sphIntersect(self, ro, rd, ra):
        b = dot(ro, rd)
        c = dot(ro, ro) - ra * ra
        h = b * b - c
        if h < 0.0:
            return vec2(-1.0, -1.0)
        else:
            h = math.sqrt(h)
            return vec2(-b - h, -b + h)

    def boxIntersection(self, ro, rd, pos1, pos2):
        pass

    def lineHit(self, vPosLine, vDirectionLine):
        for _ in self.bodyParts:
            hit = _.hitbox
            if hit['type'] == 'sphere':
                it = self.sphIntersect(vPosLine-_['pos'], vDirectionLine, _['rad'])
                it = it[0] if it[0] > 0.0 else it[1]
                if it < 0.0:
                    return False
                else:
                    itPos = vPosLine + vDirectionLine * it
                    s = distance(vPosLine, itPos)
                    return s
            elif hit['type'] == 'cube':
                return False



# boxN;
# vec3
# boxPos = vec3(3.0, 1.0, -0.001);
# it = boxIntersection(ro - boxPos, rd, vec3(1.0), boxN);
# if (it.x > 0 & & it.x < it.y) {it.x = it.x;} else {it.x = it.y;};
# if (it.x > 0.0 & & it.x < minIt.x) {
# minIt = it;
# n = boxN;
# col = vec4(0.9, 0.2, 0.2, -0.5);
# }




import time

def vecLen(v): return (v[0]**2+v[1]**2+v[2]**2)**0.5

def normVec(v):
    nl = vecLen(v)
    if nl != 0:
        nl = 1/nl
        return [v[0]*nl, v[1]*nl, v[2]*nl]
    else:
        return [0, 0, 0]

class F:
    def __init__(self, v, m):
        self.v = v
        self.m = m
        self.f = [i*m for i in v]
    
    def sum(self, f, c=1):
        f1 = [self.f[i] + f.f[i]*c for i in range(3)]
        if f1 !=[0, 0, 0]:
            m = vecLen(f1)
            v = normVec(f1)
            return F(v, m)
        else:
            return F([0, 0, 0], 0)
    
    def mul(self, f):
        f1 = [self.f[i]*f.f[i] for i in range(3)]
        if f1 !=[0, 0, 0]:
            m = vecLen(f1)
            v = normVec(f1)
            return F(v, m)
        else:
            return F([0, 0, 0], 0)
    
    def roundVec(self): 
        self.f = [round(i, 3) for i in self.f]
        if self.f != [0, 0, 0]:
            m = vecLen(self.f)
            v = normVec(self.f)
            return F(v, m)
        else:
            return F([0, 0, 0], 0)

def mul(f1, f2):
    f = [f1.f[i]*f2.f[i] for i in range(3)]
    if f !=[0, 0, 0]:
        m = vecLen(f1)
        v = normVec(f1)
        return F(v, m)
    else:
        return F([0, 0, 0], 0)



import glm
import pygame as pg


class Camera:
    def __init__(self, app, position=(0,0,4), yaw=-90, pitch=0):
        self.app = app
        self.rotMouse = [0, 0]
        self.visible = True
        self.cinematog = False
        self.pcm = {True:[0.02, 0.94], False:[0.15, 0.5]}
        self.aspect_ratio = app.size[0] / app.size[1]
        self.FOV = 60
        self.NEAR = 0.1
        self.FAR = 1000
        self.position = glm.vec3(position)
        self.up = glm.vec3(0, 1, 0)
        self.right = glm.vec3(1, 0, 0)
        self.forward = glm.vec3(0, 0, -1)
        self.yaw = yaw
        self.pitch = pitch
        self.px, self.pn = 90, -90
        self.view = 300
        self.gvm()
        self.gpm()

    def rot(self, rel):
        self.rotMouse[0] += rel[0] * self.pcm[self.cinematog][0]
        self.rotMouse[1] += rel[1] * self.pcm[self.cinematog][0]

    def ucv(self):
        yaw, pitch = glm.radians(self.yaw), glm.radians(self.pitch)
        self.forward.x = glm.cos(yaw) * glm.cos(pitch)
        self.forward.y = glm.sin(pitch)
        self.forward.z = glm.sin(yaw) * glm.cos(pitch)
        self.forward = glm.normalize(self.forward)
        self.right = glm.normalize(glm.cross(self.forward, glm.vec3(0,1,0)))
        self.up = glm.normalize(glm.cross(self.right, self.forward))

    def update(self):
        self.ucv()
        self.gvm()
        self.gpm()
        self.yaw += self.rotMouse[0]
        self.pitch -= self.rotMouse[1]
        self.pitch = max(self.pn, min(self.px, self.pitch))
        self.rotMouse = [i * self.pcm[self.cinematog][1] for i in self.rotMouse]

    def vis(self): self.visible = not self.visible; pg.mouse.set_visible(self.visible); pg.event.set_grab(not(self.visible));
    def cin(self): self.cinematog = not self.cinematog
    def gvm(self): self.m_view = glm.lookAt(self.position, self.position + self.forward, self.up)
    def gpm(self): self.m_proj = glm.perspective(glm.radians(self.FOV), self.aspect_ratio, self.NEAR, self.FAR)
